"""Attrify decorates a custom class leaving a new class subclassing the
decorated class, and with the AttriClass.__dict__ as the class body."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from attribox import AttriClass
from vistutils.waitaminute import typeMsg


class Attrify:
  """Attrify decorates a custom class leaving a new class subclassing the
  decorated class, and with the AttriClass.__dict__ as the class body."""

  __pos_args__ = None
  __key_args__ = None

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the decorator."""
    self.__pos_args__ = args
    self.__key_args__ = kwargs

  def __call__(self, cls: type) -> type:
    """Returns a new class subclassing"""
    name = cls.__name__
    bases = (cls,)
    namespace = {**AttriClass.__dict__, '__pos_args__': self.__pos_args__,
                 '__key_args__'                       : self.__key_args__}
    return type(name, bases, namespace)



"""The example package provides a simple example of how the 'attribox'
package can be used. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._point import Point
from ._rectangle import Rectangle



"""Example test script."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from example import Rectangle

if __name__ == '__main__':
  r = Rectangle(0, 0, 69, 420)
  print(r)
  r.topLeft.x = 77
  print(r)



"""Rectangle provides a class representation of a plane rectangle."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from attribox import AttriBox
from example import Point


class Rectangle:
  """This rectangle now uses AttriBox to define topLeft and bottomRight
  corners. """

  __field_name__ = None
  __field_owner__ = None

  topLeft = AttriBox[Point](0, 0)
  bottomRight = AttriBox[Point](0, 0)

  def __init__(self, *args) -> None:
    print('%s start of __init__' % self.__class__.__name__)
    intArgs = []
    pointArgs = []
    for arg in args:
      if isinstance(arg, int):
        intArgs.append(arg)
      elif isinstance(arg, Point):
        pointArgs.append(arg)
    if len(pointArgs) == 2:
      self.topLeft.x = pointArgs[0].x
      self.topLeft.y = pointArgs[0].y
      self.bottomRight.x = pointArgs[1].x
      self.bottomRight.y = pointArgs[1].y
    elif len(intArgs) == 4:
      self.bottomRight.y = intArgs.pop()
      self.bottomRight.x = intArgs.pop()
      self.topLeft.y = intArgs.pop()
      self.topLeft.x = intArgs.pop()
    clsName = self.__class__.__name__
    print('End of %s.__init__' % (clsName,))

  def getName(self) -> str:
    """Getter-function for instance name. If __set_name__ has run, the
    name received is included here."""
    if self.__field_name__ is None:
      return self.__class__.__name__
    return '%s.%s' % (self.__field_owner__.__name__, self.__field_name__)

  def __str__(self, ) -> str:
    """String representation of the rectangle. """
    return '%s(%s, %s)' % (self.getName(), self.topLeft, self.bottomRight)

  def area(self) -> int:
    """Returns the area of the rectangle. """
    v = self.topLeft - self.bottomRight
    return abs(v.x * v.y)



"""Point provides a class representation of a two-dimensional point. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from attribox import AttriBox


class Point:
  """This example class uses AttriBox to implement coordinates"""

  __field_name__ = None
  __field_owner__ = None

  x = AttriBox[int](0)
  y = AttriBox[int](0)

  def __set_name__(self, owner: type, name: str) -> None:
    """Sets the name of the field. This method runs only when the instance
    of the class is itself used as a descriptor."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __init__(self, x: int = 0, y: int = 0) -> None:
    print('Point.__init__ at: (%d, %d)' % (x, y))
    self.x = x
    self.y = y

  @x.ONGET
  def _notifyGet(self, value: int) -> None:
    """This method notifies when an AttriBox calls __get__"""
    # value = maybe(value, 0)
    # print('%s.x get -> %d' % (self.getName(), value))

  @x.ONSET
  def _notifySet(self, oldVal: int, newVal: int) -> None:
    """This method notifies when an AttriBox calls __set__"""
    # newVal = maybe(newVal, 0)
    # oldVal = maybe(oldVal, 0)
    # print('%s.x set from %d to %d' % (self.getName(), oldVal, newVal))

  def __add__(self, other: Point) -> Point:
    """Implementation of the addition operator."""
    return Point(self.x + other.x, self.y + other.y)

  def __sub__(self, other: Point) -> Point:
    """Implementation of the subtraction operator."""
    return Point(self.x - other.x, self.y - other.y)

  def getFieldName(self) -> str:
    """Getter-function for the field name. """
    if self.__field_name__ is None:
      return self.__class__.__name__
    return self.__field_name__

  def getOwnerName(self) -> str:
    """Getter-function for the field owner. """
    if self.__field_owner__ is None:
      return self.__class__.__name__
    return self.__field_owner__.__name__

  def getName(self) -> str:
    """Getter-function for instance name. If __set_name__ has run, the
    name received is included here."""
    if self.__field_name__ is None:
      return self.__class__.__name__
    ownerName = self.getOwnerName()
    fieldName = self.getFieldName()
    return '%s.%s' % (ownerName, fieldName)

  def __str__(self, ) -> str:
    """String representation of the point. """
    return '(%d, %d)' % (self.x, self.y)

  def __repr__(self) -> str:
    """Code representation"""
    return '%s(%d, %d)' % (self.__class__.__name__, self.x, self.y)
